'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.3.4

version: 1.3.4

full_version: 1.3.4

git_revision: dc0c75e9e1e0122bf839af0ecb9d5ffee0d45a4f

release: True

'''
short_version: str = '1.3.4'
version: str = '1.3.4'
full_version: str = '1.3.4'
git_revision: str = 'dc0c75e9e1e0122bf839af0ecb9d5ffee0d45a4f'
release: bool = True
if not release:
    version = full_version
